# This subroutine can be appended to qmail-scanner-queue.pl.
# It will restrict Q-S to stop sending alert after alert to the
# same email address if a quarantine event occurs.
#
# Simply append this subroutine to the bottom of qmail-scanner-queue.pl
# and change the email_sender subroutine so that:
#
#  if (!&is_unreplyable_email('sender') && &notify_addr('sender')) {
#    &debug("e_s: sending quarantine report via: $qmailinject to sender address ($returnpath)");
#    print SM "To: $returnpath\n";
#  }
#
# ... becomes ...
#
#
#  if (!&is_unreplyable_email('sender') && &notify_addr('sender')) {
#    if (!&generate_sender_db($returnpath,$remote_smtp_ip)) {
#      &debug("e_s: sending quarantine report via: $qmailinject to sender address ($returnpath)");      
#	print SM "To: $returnpath\n";
#    } else {
#      &debug("e_s: skipping quarantine report to sender due to excessive previous reports");
#    }
#  }
#
# There are three variables you can alter: $MAX_ALERT_FREQ, $FLUSH_ALERT_DB 
# and $IP_ADDR_CHECK.
#
# Once installed, every time Q-S intends to send an alert to a sender, it 
# will first check the sender cache DB. If the senders address (or IP address)
# is > $MAX_ALERT_FREQ, then no alert is send to the sender. Also, if
# any records in the DB are > $FLUSH_ALERT_DB days old, then they are flushed
# out.

sub generate_sender_db {
  my ($email,$ipaddr)=@_;
  
  &debug("g_s_db: this has been called");
  return if ($email =~ /^($QUARANTINE_CC|$V_FROM)/i);
  use DB_File;

  my ($addr_cache)= "address_cache";
  my ($status)=0;
  my ($MAX_ALERT_FREQ)=3; # Max of 3 alerts to a user before stopping alerts
  my ($FLUSH_ALERT_DB)=30; # flush old entries after 30 days
  my ($IP_ADDR_CHECK) = 1; #set to '0' to disable IP address checks
  my ($addr,%alert_array);
  my ($num,$num1,$time,$time1)='';

  tie (%alert_array, "DB_File", "$addr_cache.db", O_CREAT|O_RDWR, 0640, $DB_HASH)  || return; #won't check for errors, life goes on if this fails...
  ($num,$time)=split(/:/,$alert_array{$email},2);
  if (!$num) { $num=0;$time=time};
  &debug("g_s_db: currently $email has $num entries");
  $num++;
  if (time > ($time + $FLUSH_ALERT_DB*24*3600)) {
    delete $alert_array{$email};
    &debug("g_s_db: expiring entry for $email as $time < ".time);
    $num=1;
  } elsif ($num > $MAX_ALERT_FREQ ) {
    $status=1;
  }
  &debug("g_s_db: setting $email $num at time $time");
  $alert_array{$email}="$num:$time";
  if ($IP_ADDR_CHECK) {
    ($num1,$time1)=split(/:/,$alert_array{$ipaddr},2);
    if (!$num1) { $num1=0;$time1=time};
    &debug("g_s_db: currently $ipaddr has $num1 entries");
    $num1++;
    if (time > ($time1 + $FLUSH_ALERT_DB*24*3600)) {
      delete $alert_array{$ipaddr};
      &debug("g_s_db: expiring entry for $ipaddr");
      $num1=1;
    } elsif ($num1 > $MAX_ALERT_FREQ && $ipaddr ne "" ) {
      $status=1;
    }
    &debug("g_s_db: setting $ipaddr to $num1 at time $time");
    $alert_array{$ipaddr}="$num1:$time1";
  }
  untie %alert_array;
return $status;
}
