#################################################
# Subroutines added by ST
#################################################

sub minidebug {
  my $dnowtime = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime(time));
  print LOG "$dnowtime:$nprocess: ",@_,"\n" if ($MINIDEBUG && !$DEBUG);
}

sub close_log {
  ($sec,$min,$hour,$mday,$mon,$year) = localtime(time);

  &debug("--- all finished. Total of ",tv_interval ($start_time, [gettimeofday])," secs");
  &minidebug("------ Process $nprocess finished. Total of ",tv_interval ($start_time, [gettimeofday])," secs");
  close(LOG);
}

sub reject_email {
  my ($exit_string,$exit_code)=@_;
  $exit_code=111 if (!$exit_code);

  # st: tell qmail-smtpd why the message is rejected,
  # so it can be written to the qmail-smtpd log
  warn "$V_HEADER-$VERSION: $exit_string\n" if ($MINIDEBUG <= 2);
  warn "$nppid QS-$VERSION: $exit_string\n" if ($MINIDEBUG > 2);

  &debug("r_e: $V_HEADER-$VERSION: $exit_string");
  &minidebug("r_e: $V_HEADER-$VERSION: $exit_string") if ($MINIDEBUG <= 2);
  &minidebug("r_e: QS-$VERSION: $exit_string") if ($MINIDEBUG > 2);

  &cleanup;

  # st: support for qmail-queue-custom-error.patch
  open(FD,">&$REJECT_FD");
  print FD "D$exit_string";
  close(FD);

  $elapsed_time = tv_interval ($start_time, [gettimeofday]);
  &log_event;
  &close_log;
  exit $exit_code;
}

##############################################
# st: SETTINGS PER DOMAIN routines
##############################################

sub start_scanners {
  my($e_sender,$f_recips,$msg)=@_;
  $sa_rcpt='0';

  # Now, start the scanners!
  &init_scanners if ($scanner_array[0] ne "none");

  # st: if the message is marked to delete skip the mailing routines
  if (!$del_message) {
    if (($quarantine_event || $quarantine_spam) && ($scanner_array[0] ne "none")) {
      &debug("unsetting TCPREMOTEIP env var");
      delete $ENV{'TCPREMOTEIP'};
      #Reset locale back to original
      $ENV{'LC_ALL'}=$orig_locale;

      if ($sa_forward ne "" && $quarantine_event =~/spam/i && $description !~/potential virus/i) {
         if ($sa_fwd_verbose) {
            $sa_hdr_report='1' if ($sa_alt && $sa_report);
            &qmail_parent_check;
            &qmail_requeue($e_sender,"T$sa_forward\0\0",$msg);
         } else {
            open (SF,"$qmailinject -f$returnpath $sa_forward < $msg|")||&error_condition("cannot run $qmailinject -f$returnpath $sa_forward < $msg - $!");
            close SF ;
         }
         # st: forward the messages just once..
         $sa_rcpt='0';
         $sa_forward='';
      }
      ## st: This code is from qs-2.00, I have to check...
      #is this a greylist event?
      if ($quarantine_event=~/gr[ae]ylist/i ) {
        #This text will only be seen by those using the "custom-error"
        #patch. Others will just get a general "qq" temp failure msg.
        &log_event;
        print STDERR "Z$quarantine_event";
        &cleanup;
        &close_log;
        exit 82;
      }else{    
        &email_quarantine_report;
      }
      ##
    } else {
      &qmail_parent_check;
      &qmail_requeue($e_sender,$f_recips,$msg); 
    }
  } else {
    $elapsed_time = tv_interval ($start_time, [gettimeofday]);
  }
}

sub sa_defaults {
  $sa_subject=$sa_subject_site;
  $sa_quarantine=$sa_quarantine_site;
  $sa_delta=$sa_delta_site;
  $sa_delete=$sa_delete_site;
  $sa_reject=$sa_reject_site;
  $sa_forward=$sa_forward_site;
  $sa_fwd_verbose=$sa_fwd_verbose_site;
  $sa_hdr_report=$sa_hdr_report_site;
  $smaildir=$smaildir_site;
}

sub settings_pd {
  my ($match_hdr,$match_rcpt,$domain_settings)=@_;
  my ($scanners_rcpt);

  ($scanners_rcpt,$sa_subject,$sa_quarantine,$sa_delta,$sa_delete,$sa_reject,$sa_forward,$sa_fwd_verbose,$sa_hdr_report,$smaildir)=split(/'/,$domain_settings);
  $sa_subject="" if ($sa_subject eq "none");
  $smaildir=untaint($smaildir);  # st: suggested by P-O Yliniemi <peo - bsd-guide.net>
  $sa_forward=untaint($sa_forward); # st: sa_forward must be untainted too, thanks to Tomas Charvat <tc - excello.cz>

  &debug("s_p_d: $match_hdr match '$match_rcpt', settings '$sa_subject,$sa_quarantine,$sa_delta,$sa_delete,$sa_reject,$sa_forward,$sa_fwd_verbose,$sa_hdr_report,$smaildir'");

  @scanner_array=split(/,/,$scanners_rcpt);

  &debug("s_p_d: $match_hdr match '$match_rcpt', scanners '$scanners_rcpt'");
  &minidebug("s_p_d: $match_hdr match '$match_rcpt', scanners '$scanners_rcpt'") if ($match_hdr !~ /m_rcpt/);
}

sub settings_p_d {
  my (%domain_settings,%seen,$scanners_array,$scanners_rcpt,$domain_settings);

  &debug("s_p_d: reading from $settings_per_domain.db");
  tie (%domain_settings,'DB_File',"$settings_per_domain.db",O_RDONLY, 0600, $DB_HASH) || &error_condition("cannot open $settings_per_domain.db - $!");

  # Check if we have a match within the database
  # Check order:
  # 1) return-path
  # 2) domain-return-path
  # 3) for each recipient: recipient, domain-recipient
  if ((exists $domain_settings{$returnpath}) && $QS_RELAYCLIENT) {
     &settings_pd ("return-path",$returnpath,$domain_settings{$returnpath});
  }
  elsif ((exists $domain_settings{$domain_returnpath}) && $QS_RELAYCLIENT) {
     &settings_pd ("domain-return-path",$domain_returnpath,$domain_settings{$domain_returnpath});
  }
  elsif ($one_recip && (exists $domain_settings{$one_recip})) {
     &settings_pd ("rcpt",$one_recip,$domain_settings{$one_recip});
  }
  elsif ($one_recip && (exists $domain_settings{$domain_one_recip})) {
     &settings_pd ("domain_rcpt",$domain_one_recip,$domain_settings{$domain_one_recip});
  }
  elsif (!$one_recip) {
     &debug("s_p_d: we have multiple recipient, checking each of them");
     &minidebug("s_p_d: we have multiple recipient, checking each of them");
     my @mrecips=split(',',$recips);
     my $mrcpt='';
     my $domain_mrcpt='';
     my %m_rcpt;
     foreach $mrcpt(@mrecips) {
       $mrcpt=tolower($mrcpt);
       $domain_mrcpt=$mrcpt;
       $domain_mrcpt=~ s/^(.*)\@(.*)$/$2/;
       if (exists $domain_settings{$mrcpt}) {
          &settings_pd ("m_rcpt",$mrcpt,$domain_settings{$mrcpt});
       }
       elsif (exists $domain_settings{$domain_mrcpt}) {
          &settings_pd ("domain-m_rcpt",$domain_mrcpt,$domain_settings{$domain_mrcpt});
       } else {
          @scanner_array=@scanners_default;
          &sa_defaults;
       }
       @scanner_array=&check_scanners(@scanner_array);
       $scanners_rcpt=join(',',@scanner_array);
       $domain_settings="$scanners_rcpt'$sa_subject'$sa_quarantine'$sa_delta'$sa_delete'$sa_reject'$sa_forward'$sa_fwd_verbose'$sa_hdr_report'$smaildir";
       $m_rcpt{$mrcpt}=$domain_settings;
     }
     untie %domain_settings;
     while( ($one_recip,$scanners_array)=each %m_rcpt) {
       &settings_pd ("rcpt",$one_recip,$scanners_array);
       &start_scanners($env_returnpath,"T$one_recip\0\0","$scandir/$wmaildir/new/$file_id");
       # st: maybe I had to change this if I will ever do 'sa' per user config...
       # if an user on a multiples recipients mail has a very low sa_delete... It could
       # be rare, but it could be. What to do?
       # If sa_hits doesn't exist, the mail has a virus marked to delete,
       # but if the mail was rejected this check won't be reached...
       last if ($del_message == 1);
     }
     return;
  } else {
     @scanner_array=@scanners_default;
     &sa_defaults;
     &debug("s_p_d: no match, default sa_settings '$sa_quarantine,$sa_delta,$sa_delete,$sa_reject,$sa_forward,$sa_fwd_verbose,$sa_hdr_report,$smaildir'");
     &debug("s_p_d: no match, falling to settings_default");
     &minidebug("s_p_d: no match, falling to settings_default");
  }
  # if no multiples recipients
  untie %domain_settings;
  @scanner_array=&check_scanners(@scanner_array);
  &start_scanners($env_returnpath,$env_recips,"$scandir/$wmaildir/new/$file_id");
}

sub generate_spd {
  my ($line,$count,%domain_settings,$match_rcpt,$scanners_rcpt,@scanners_rcpt_array,%seen);
  my ($domain_settings,$sa_subject_ignore);

  print "\n  Generating $settings_per_domain.db\n\n";

  unlink ("$settings_per_domain.db.tmp");
  tie (%domain_settings,'DB_File',"$settings_per_domain.db.tmp",O_CREAT|O_RDWR,0640,$DB_HASH) || &error_condition("cannot open for write $settings_per_domain.db.tmp - $!");

  open(SPD, "<$settings_per_domain.txt") || &error_condition("cannot read $settings_per_domain.txt - $!");

  while (<SPD>) {
    $line++;
    next if (/^\#|^\s.*$/);    # Ignore lines starting with # or spaces
    next if (!(/:/));          # Ignore lines doesn't contain a ':'
    # if (/\;|\!/) {
    if (/\;/) {
       print "d_w: line $line contains an invalid char, SKIP\n";
       next;
    }
    chomp;
    # sa_subject could has spaces ... (from  P-O Yliniemi)
    $sa_subject = (split(/'/,$_))[1];
    s/\s|\t//g;
    ($match_rcpt,$domain_settings)=split(/:/,$_,2);
    $match_rcpt=tolower("$match_rcpt");
    # $domain_settings=tolower("$domain_settings");
    ($scanners_rcpt,$sa_subject_ignore,$sa_quarantine,$sa_delta,$sa_delete,$sa_reject,$sa_forward,$sa_fwd_verbose,$sa_hdr_report,$smaildir)=split(/'/,$domain_settings);

    if (exists $domain_settings{$match_rcpt}) {
       print "  d_w: duplicated value '$match_rcpt' at line $line, SKIP \n";
       next;
    }

    $sa_subject=$sa_subject_site if (!$sa_subject);
    $sa_quarantine=$sa_quarantine_site if (!$sa_quarantine && $sa_quarantine ne "0");
    $sa_delta=$sa_delta_site if (!$sa_delta && $sa_delta ne  "0");
    $sa_delete=$sa_delete_site if (!$sa_delete && $sa_delete ne "0");
    $sa_reject=$sa_reject_site if (!$sa_reject && $sa_reject ne "0");
    $sa_forward=$sa_forward_site if (!$sa_forward);
    $sa_fwd_verbose=$sa_fwd_verbose_site if (!$sa_fwd_verbose && $sa_fwd_verbose ne "0");
    $sa_hdr_report=$sa_hdr_report_site if (!$sa_hdr_report && $sa_hdr_report ne "0");
    $smaildir=$smaildir_site if (!$smaildir);

    # Control the values of sa_delete and sa_quarantine
    if ($sa_delete && ($sa_quarantine>$sa_delete)) {
       print "  d_w: WARNING, sa_delete lower than sa_quarantine, for address '$match_rcpt' at line $line\n";
       print "       resetting sa_delete to '0', spam could be quarantined, but not deleted for this address\n";
       $sa_delete='0';
    }

    # Let check if the scanner are really installed,
    # change 'sa' and 'ps' for the correct name, and
    # add _scanner to the AVs scanners

    @scanners_rcpt_array=split(/,/,$scanners_rcpt);
    foreach (@scanners_rcpt_array) {
       s/^sa$/spamassassin/;
       s/^ps$/perlscan/;
       s/^perlscanner$/perlscan/;
       s/^(.*)$/$1_scanner/ if((!/spamassassin/) && (!/_scanner/) && (!/^none$/));
    }

    # Check if the scanners are installed
    @scanners_rcpt_array=&check_scanners(@scanners_rcpt_array);

    $scanners_rcpt = join(',',@scanners_rcpt_array);

    # Check if at least we have one valid scanner

    if (@scanners_rcpt_array==0) {
       print "  d_w: There are no valid scanners for address '$match_rcpt' at line $line, SKIP\n";
       next;
    }
    $count++;

    $domain_settings="$scanners_rcpt'$sa_subject'$sa_quarantine'$sa_delta'$sa_delete'$sa_reject'$sa_forward'$sa_fwd_verbose'$sa_hdr_report'$smaildir";
  
    $domain_settings{$match_rcpt}=$domain_settings;
  }
  close(SPD);
  untie %domain_settings;
  rename( "$settings_per_domain.db.tmp", "$settings_per_domain.db" );
  print "\n  Read $line lines, got $count entries\n\n";
  if (!$settings_pd) {
     print "\n  WARNING: settings_per_domain is not enabled\n\n  The database has been generated but\n";
     print "  it won't be used until 'settings_per_domain' will be enabled\n\n";
  }
}

sub read_spd {
  # st: display the database sorted by domains.

  my ($count,%domain_settings,$scanners_rcpt,$TXT,$spd_orig);
  my (%sorted,$userpart,$domainpart,$last_domain,$fs);
  $count=0;

  if ($opt_d) {
     $fs="\t";
     print "\n#  Reading from $settings_per_domain.db\n#\n";
     $TXT="STDOUT";
  } else {
     $fs="\n#";
     $spd_orig = "$settings_per_domain.txt.";
     $spd_orig .= sprintf "%02d%02d%02d.%02d%02d%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec;
     print "\n  Sorting file '$settings_per_domain.txt'\n\n  A copy of the current settings will be saved in:\n";
     print "  $spd_orig\n\n";
     if ( (stat("$settings_per_domain.txt"))[9] > (stat("$settings_per_domain.db"))[9] ) {
        print "\n  WARNING: file '$settings_per_domain.txt' is newer\n";
        print "           than database '$settings_per_domain.db'\n";
        print "           this could be wrong or not, anyway a copy of the current\n";
        print "           'txt file' has been saved, see above.\n\n";
     }
     rename("$settings_per_domain.txt", "$spd_orig");
     open(SPD, ">$settings_per_domain.txt") || &error_condition("cannot open for write $settings_per_domain.txt - $!");
     print SPD "#\n# File settings_per_domain.txt sorted by qmail-scanner-st\n#\n";
     $TXT="SPD";
  }

  print $TXT "#  Read the documetation at:\n";
  print $TXT "#  http://toribio.apollinare.org/qmail-scanner/settings_per_domain.html\n#\n";

  if ($opt_s) {
     print $TXT "\n######### WIDE SITE SETTINGS\n";
     print $TXT "# scanners_installed = @scanners_installed\n";
     print $TXT "# scanners_default   = @scanners_default\n";
     print $TXT "# sa_subject_site    = '$sa_subject_site'\n";
     print $TXT "# sa_quarantine_site = $sa_quarantine_site$fs sa_delta_site      = $sa_delta_site\n";
     print $TXT "# sa_delete_site     = $sa_delete_site$fs sa_reject_site     = $sa_reject_site\n";
     print $TXT "# sa_forward_site    = '$sa_forward_site'$fs sa_fwd_verbose_site= $sa_fwd_verbose_site\n";
     print $TXT "# sa_hdr_report_site = $sa_hdr_report_site$fs smaildir_site      = $smaildir_site\n\n";
  }

  tie (%domain_settings,'DB_File',"$settings_per_domain.db",O_RDONLY, 0600, $DB_HASH) || &error_condition("cannot open for read $settings_per_domain.db - $!");;

  # st: let sort the match_rpt
  foreach (keys %domain_settings) {
     if ( $_ =~ /\@/) {
        ($userpart,$domainpart) = split (/\@/,$_);
        $sorted{"$domainpart.$userpart"} = $_;
     } else {
        $sorted{$_} = $_;
     }
  }

  foreach(sort keys %sorted) {
     $count++;
     ($userpart,$domainpart) = split (/\@/,$sorted{$_});
     if ( $sorted{$_} !~ /\@/ ) {
        print $TXT "\n######### DOMAIN\t'$sorted{$_}'\n" ;
        $last_domain=$domainpart=$sorted{$_};
     }
     print $TXT "\n######### DOMAIN\t'$domainpart'\n" if ( $domainpart ne $last_domain );
     $last_domain=$domainpart;
     ($scanners_rcpt,$sa_subject,$sa_quarantine,$sa_delta,$sa_delete,$sa_reject,$sa_forward,$sa_fwd_verbose,$sa_hdr_report,$smaildir)=split(/'/,$domain_settings{$sorted{$_}});
     print $TXT "\n## $count. Settings for\t'$sorted{$_}'\n" if ($opt_d) ;
     print $TXT "$sorted{$_} : $domain_settings{$sorted{$_}}\n";
     if ($opt_d) {
        print $TXT "\n# scanners      = $scanners_rcpt\n";
        print $TXT "# sa_subject    = '$sa_subject'\n";
        print $TXT "# sa_quarantine = $sa_quarantine\tsa_delta       = $sa_delta\n";
        print $TXT "# sa_delete     = $sa_delete\tsa_reject      = $sa_reject\n";
        print $TXT "# sa_forward    = '$sa_forward'\tsa_fwd_verbose = $sa_fwd_verbose\n";
        print $TXT "# sa_hdr_report = $sa_hdr_report\tsmaildir       = $smaildir\n\n";
     }
  }

  print $TXT "\n######### WIDE SITE SETTINGS\n";
  print $TXT "# scanners_installed = @scanners_installed\n";
  print $TXT "# scanners_default   = @scanners_default\n";
  print $TXT "# sa_subject_site    = '$sa_subject_site'\n";
  print $TXT "# sa_quarantine_site = $sa_quarantine_site$fs sa_delta_site      = $sa_delta_site\n";
  print $TXT "# sa_delete_site     = $sa_delete_site$fs sa_reject_site     = $sa_reject_site\n";
  print $TXT "# sa_forward_site    = '$sa_forward_site'$fs sa_fwd_verbose_site= $sa_fwd_verbose_site\n";
  print $TXT "# sa_hdr_report_site = $sa_hdr_report_site$fs smaildir_site      = $smaildir_site\n\n";
  if ($opt_d) {
     print $TXT "# Run '/var/qmail/bin/qmail-scanner-queue.pl -p' to generate the db\n";
     print $TXT "# If you have redirect the output of this command to settings_per_domain.txt\n";
     print $TXT "\n# d_w: total of $count entries found\n\n\n";
  }
  if (!$settings_pd) {
     print "\n  WARNING: settings_per_domain is not enabled\n\n  The database won't be used\n";
     print "  until 'settings_per_domain' will be enabled\n\n";
  }

  close(SPD) if ($opt_s);
  untie %domain_settings;
}


sub check_scanners {
  # Check against the installed scanners
  my @scanners_to_check=@_;
  return @scanners_to_check if ($scanners_to_check[0] eq "none");
  my %seen=();
  foreach (@scanners_installed) {
     $seen{$_}=1;
  }

  @scanners_to_check=grep($seen{$_},@scanners_to_check);
  return @scanners_to_check;
}

sub untaint {
  # st: suggested by P-O Yliniemi <peo - bsd-guide.net>
  my($var) = @_;
  if ($var =~ /^(.*)$/) {
     $var = $1;
  }
  return  $var;
}

#################################################
# END of subroutines added by ST
#################################################

